""" Minesweeper util """

from os.path import join, dirname
import pygame
import copy

class Spritesheet:
    """ Spritesheet """
    def __init__(self, filename):
        try:
            self.sheet = pygame.image.load(filename).convert()
        except pygame.error as message:
            print("Unable to load spritesheet image: %s" % filename)
            raise SystemExit(message)

    def image_at(self, rectangle, colorkey=None):
        "Loads image from x,y,x+offset,y+offset"
        rect = pygame.Rect(rectangle)
        image = pygame.Surface(rect.size).convert()
        image.blit(self.sheet, (0, 0), rect)
        if colorkey is not None:
            if colorkey is -1:
                colorkey = image.get_at((0, 0))
            image.set_colorkey(colorkey, pygame.RLEACCEL)
        return image

    def images_at(self, rects, colorkey=None):
        "Loads multiple images, supply a list of coordinates"
        return [self.image_at(rect, colorkey) for rect in rects]

    def load_strip(self, rect, image_count, colorkey=None):
        "Loads a strip of images and returns them as a list"
        tups = [(rect[0]+rect[2]*x, rect[1], rect[2], rect[3])
                for x in range(image_count)]
        return self.images_at(tups, colorkey)

    def load_grid(self, grid, colorkey=None):
        "load grids of images and returns them as a list"
        rects = []
        size_x = self.sheet.get_width()  // grid[0]
        size_y = self.sheet.get_height() // grid[1]
        for pos_y in range(0, self.sheet.get_height(), size_y):
            for pos_x in range(0, self.sheet.get_width(), size_x):
                rects.append((pos_x, pos_y, size_x, size_y))
        return self.images_at(rects, colorkey)


class Tile(Spritesheet):
    """ Tile """
    def  __init__(self, api, color=True):
        if color:
            super().__init__(join(dirname(__file__), "color_tiles.png"))
        else:
            super().__init__(join(dirname(__file__), "bw_tiles.png"))
        images = self.load_grid((8, 2))
        self.images = {
            api.encoder.ascii_tile      : images[0],
            api.encoder.ascii_mine      : images[5],
            'f'  : images[2],
            '?'  : images[3],
            '*?' : images[4],
            '*x' : images[6],
            'fx' : images[7],

            # nums
            '0' : images[1],
            '1' : images[8],
            '2' : images[9],
            '3' : images[10],
            '4' : images[11],
            '5' : images[12],
            '6' : images[13],
            '7' : images[14],
            '8' : images[15],
        }
        self.width = images[0].get_width()
        self.height = images[0].get_height()

    def __getitem__(self, tile):
        return self.images[tile]


class Face(Spritesheet):
    """ Face """
    def __init__(self, color=True):
        if color:
            super().__init__(join(dirname(__file__), "color_face.png"))
        else:
            super().__init__(join(dirname(__file__), "bw_tiles.png"))
        images = self.load_grid((5, 1))
        self.images = {
            ':)'  : images[0],
            '*:)' : images[1],
            ':o'  : images[2],
            'B)'  : images[3],
            'x('  : images[4],
        }
        self.width = images[0].get_width()
        self.height = images[0].get_height()
    def __getitem__(self, face):
        return self.images[face]



class Score(Spritesheet):
    """ Score """
    def __init__(self, color=True):
        if color:
            super().__init__(join(dirname(__file__), "color_scoreboard.png"))
        else:
            super().__init__(join(dirname(__file__), "bw_scoreboard.png"))

        images = self.load_grid((10, 1))
        self.images = {
            '0' : images[0],
            '1' : images[1],
            '2' : images[2],
            '3' : images[3],
            '4' : images[4],
            '5' : images[5],
            '6' : images[6],
            '7' : images[7],
            '8' : images[8],
            '9' : images[9],
        }
        self.width = images[0].get_width()
        self.height = images[0].get_height()
    def __getitem__(self, number):
        return self.images[str(number)]



#!/usr/bin/env python
""" Minesweeper """

try:
    import minesweeper
except ImportError:
    from . import minesweeper

class Solver:
    def __init__(self, API : minesweeper.API):
        self.api = API

if __name__ == "__main__":
    s = Solver(minesweeper.API(12, 12, 10))


# -*- coding: utf-8 -*-
""" Minesweeper """
import random
import time

def valid_arguments(fn):
    def init(self, width, height, mines):
        try:
            w, h, m = int(width), int(height), int(mines)
            if not ((w >= 10 and h >= 10) and (w <= 100 and h <= 100)):
                raise ValueError('Arguments "width" and "height" must range [10, 100].')
            if m < 0 and m > ((h * w) - 9):
                raise ValueError('Arguments "mines" must range (0, (h * w) - 9]')
        except TypeError:
            raise TypeError("Arguments must be a number")
        return fn(self, w, h, m)
    return init

def valid_attempt(fn):
    def is_valid_attempt(self, y, x):
        if self._is_game_over:
            return []
        if not self.tile_valid(y, x):
            return []
        if not self.encoder.is_tile(self._board[y][x]):
            return []
        if self._is_game_done:
            return []
        if self._opened == 0:
            self._sheet_init(y, x)
        return fn(self, y, x)
    return is_valid_attempt


class API:
    """ API """
    
    @valid_arguments
    def __init__(self, width, height, mines):
        """ Generates a minesweeper game.
        
        Parameters
        ----------
        width : int
            width of player's board
        height : int
            height of player's board
        mines : int
            number of mines placed randomly on the player's board
        
        Raises
        ------
        ValueError
            If arguments exceeds fixed thresholds.
        TypeError
            If arguments are not numbers
        """
        self.encoder = Encoder()
        
        self._width = width
        self._height = height
        self._mines = mines

        self._board = self._generate_board()
        self._sheet = self._generate_sheet()
        
        self._is_game_over = False
        self._is_game_done = False
        self._opened = 0
        self._timer = time.time()
        self._timer1= time.time()
    
    def game_new(self, width, height, mines):
        """ Generates a new game with different settings.
        
        Parameters
        ----------
        width : int
            width of player's board
        height : int
            height of player's board
        mines : int
            number of mines placed randomly on the player's board
        
        Raises
        ------
        ValueError
            If arguments exceeds fixed thresholds.
        TypeError
            If arguments are not numbers
        """
        self.__init__(width, height, mines)

    def game_reset(self):
        """ Resets game with the same settings except bomb locations. """
        self.__init__(self.width, self.height, self.mines)

    def _generate_board(self):
        """ Generate board matrix. """
        return [[Encoder.num_tile for _ in range(self.width)] for _ in range(self.height)]

    def _generate_sheet(self):
        """ Generates the bomb locations randomly. """
        mines = [0 for _ in range(self.width * self.height)]
        while mines.count(Encoder.num_mine) != self.mines:
            tile = random.randint(0, len(mines) - 1)
            while mines[tile] == Encoder.num_mine:
                tile = random.randint(0, len(mines) - 1)
            mines[tile] = Encoder.num_mine
        return [[mines[(y * self.width) + x] for x in range(self.width)] for y in range(self.height)]
    
    def _sheet_init(self, pos_y, pos_x):
        # Make sure the user never hits a mine on the first opened tile
        for y in [pos_y + 1, pos_y - 1, pos_y]:
            for x in [pos_x + 1, pos_x - 1, pos_x]:
                if self.tile_valid(y, x) and self.encoder.is_mine(self._sheet[y][x]):
                    i = random.randint(0, self.height - 1)
                    j = random.randint(0, self.width - 1)
                    while self.encoder.is_mine(self._sheet[i][j]) or (abs(pos_y - i) <= 1 and abs(pos_x - j) <= 1):
                        i = random.randint(0, self.height - 1)
                        j = random.randint(0, self.width - 1)
                    self._sheet[i][j] = Encoder.num_mine
                    self._sheet[y][x] = 0
        
        # make sure to reset sheet except newly created mines
        for y in range(0, self.height):
            for x in range(0, self.width):
                if not self.encoder.is_mine(self._sheet[y][x]):
                    self._sheet[y][x] = 0
        
        # calculate the numeric score of each tile 
        for y in range(0, self.height):
            for x in range(0, self.width):
                if not self.encoder.is_mine(self._sheet[y][x]):
                    for i in [y, y+1, y-1]:
                        for j in [x, x+1, x-1]:
                            if self.tile_valid(i, j) and self.encoder.is_mine(self._sheet[i][j]):
                                self._sheet[y][x] += 1
        self._timer = time.time()
    
    @valid_attempt
    def tile_open(self, y, x):
        """ Opens a tile based on given coordinate. """
        # Checks if game is not over and given tile is valid
        tile = {'value': self._sheet[y][x], 'x': x, 'y': y}
        self._board[y][x] = tile['value']
        self._opened += 1
        if (self._opened + self.mines) == (self.width * self.height):
            self._is_game_done = True
            self._timer1 = time.time()
        if self.encoder.is_mine(tile['value']):
            self._is_game_over = True
            self._timer1 = time.time()
        elif tile['value'] >= 0:
            if tile['value'] == 0:
                return self._tile_open_adjacent(y, x, [tile])
        return [tile]
    
    def _tile_open_adjacent(self, y, x, opened):
        if self.tile_valid(y, x):
            if self.encoder.is_mine(self._sheet[y][x]):
                return [{'value': Encoder.num_mine, 'x': x, 'y': y}]
            for i in [y, y+1, y-1]:
                for j in [x, x+1, x-1]:
                    if self.tile_valid(i, j) and self.encoder.is_tile(self._board[i][j]):
                        self._opened += 1
                        if (self._opened + self.mines) == (self.width * self.height):
                            self._is_game_done = True
                            self._timer1 = time.time()
                        self._board[i][j] = self._sheet[i][j]
                        opened.append({'value': self._sheet[i][j], 'x': j, 'y': i})
                        if self._sheet[i][j] == 0:
                            self._tile_open_adjacent(i, j, opened)
        return opened

    @property
    def timer(self):
        if self.is_game_over or self.is_game_done:
            return self._timer1 - self._timer if self._opened > 0 else 0.0
        return time.time() - self._timer if self._opened > 0 else 0.0

    def tile_valid(self, y, x):
        """ Check if given coordinate is within bound. """
        return True if (x >= 0 and x < self.width) and (y >= 0 and y < self.height) else False

    @property
    def is_game_over(self):
        """ Returns True if game is over False otherwise. """
        return self._is_game_over
    
    @property
    def is_game_done(self):
        """ Returns True if game is over False otherwise. """
        return self._is_game_done

    @property
    def board(self):
        return self._board

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def mines(self):
        return self._mines
    
    @property
    def sheet(self):
        return self._sheet

    @property
    def solution(self):
        return "\n".join(["".join([f"{(self.encoder[tile]):2}" for tile in row]).rstrip() for row in self._sheet])

    def __str__(self):
        return "\n".join(["".join([f"{(self.encoder[tile]):2}" for tile in row]).rstrip() for row in self._board])

class Encoder:
    ascii_tile = 't'
    ascii_mine = 'x'
    num_tile = -1
    num_mine = -2
    def __init__(self):
        self._num2ascii = {
            self.num_tile : self.ascii_tile,
            self.num_mine : self.ascii_mine,
            0 : '0',
            1 : '1',
            2 : '2',
            3 : '3',
            4 : '4',
            5 : '5',
            6 : '6',
            7 : '7',
            8 : '8',
        }
        self._ascii2num = {
            self.ascii_tile : self.num_tile,
            self.ascii_mine : self.num_mine,
            '0' : 0,
            '1' : 1,
            '2' : 2,
            '3' : 3,
            '4' : 4,
            '5' : 5,
            '6' : 6,
            '7' : 7,
            '8' : 8,
        }
    
    def ascii2num(self, string):
        try:
            return self._ascii2num[str(string)]
        except KeyError:
            raise KeyError('Argument "string" can not be encoded {}'.format(string))
    
    def num2ascii(self, integer):
        try:
            return self._num2ascii[int(integer)]
        except ValueError:
            raise ValueError('Argument "integer" must be a string')
        except KeyError:
            raise KeyError('Argument "integer" can not be encoded {}'.format(integer))

    def __getitem__(self, arg):
        if type(arg) == str:
            return self.ascii2num(arg)
        return self.num2ascii(arg)

    def is_tile(self, arg):
        if type(arg) == str:
            return True if arg == self.ascii_tile else False
        return True if arg == self.num_tile else False
    
    def is_mine(self, arg):
        if type(arg) == str:
            return True if arg == self.ascii_mine else False
        return True if arg == self.num_mine else False



#!/usr/bin/env python
""" Minesweeper """
import sys
sys.stdout = None
import pygame
sys.stdout = sys.__stdout__
import argparse
from pygame.locals import QUIT, KEYDOWN, K_ESCAPE, MOUSEBUTTONUP, MOUSEBUTTONDOWN

try:
    from .gui import Board
    from .minesweeper import API
except ImportError:
    from gui import Board
    from minesweeper import API

def init(mainf):
    """ initialize and shutsdown minesweeper"""
    def initf():
        setup()
        mainf()
        teardown()
    return initf

def setup():
    global board, resolution, screen, mines, clock, width, height
    parser = argparse.ArgumentParser(description="""
    Minesweeper

    basic        : (10x10) - 10
    intermediate : (16x16) - 40
    expert       : (30x16) - 99
    """, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("difficulty", choices=["basic", "intermediate", "expert"], help="")
    parser.add_argument("--no-color", action="store_false", help="")
    args = parser.parse_args()
    if args.difficulty == "basic":
        width, height, mines = 10, 10, 10
    elif args.difficulty == "intermediate":
        width, height, mines = 16, 16, 40
    else:
        width, height, mines = 30, 16, 99
    
    pygame.init()
    pygame.font.init()
    pygame.display.set_caption("Minesweeper")
    resolution = (width * 16, height * 16 + 40)
    screen = pygame.display.set_mode(resolution)
    clock = pygame.time.Clock()
    board = Board(API(width, height, mines), screen)

def teardown():
    pygame.quit()

@init
def main():
    """
    some doc
    """
    done = False
    rate = 0
    while not done:
        rate %= 30
        clock.tick(60)
        for event in pygame.event.get():
            if event.type == QUIT:
                done = True
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    done = True
            if event.type == MOUSEBUTTONDOWN:
                board.mouse_down(event)
                rate = 0
            if event.type == MOUSEBUTTONUP:
                board.mouse_up(event)
                rate = 0
        if rate == 0:
            board.draw(screen)
            pygame.display.flip()
        rate += 1

if __name__ == "__main__":
    main()



import pygame
from functools import reduce
try:
    import sprites
    import minesweeper
except ImportError:
    from . import sprites
    from . import minesweeper

class Board:
    def __init__(self, api, screen, color=True):
        self.board_game = BoardGame(api, y_offset=40)
        self.board_score = BoardScore(api, self.board_game.marked, 8, 8)
        self.board_face = BoardFace(api, self.game_reset)
        self.modules = [self.board_game, self.board_face, self.board_score]
        screen.fill((192, 192, 192))
        self.draw(screen)

    def create_frame(self, screen):
        pass
    
    def draw(self, screen):
        for m in self.modules:
            m.draw(screen)

    def mouse_down(self, event):
        for m in self.modules:
            m.mouse_down(event)

    def mouse_up(self, event):
        for m in self.modules:
            m.mouse_up(event)

    def game_reset(self):
        for m in self.modules:
            m.game_reset()

class BoardScore:
    def __init__(self, api, marked, x_offset=8, y_offset=8):
        self.api = api
        self.scores = sprites.Score()
        self.x_score = x_offset
        self.y_score = y_offset
        self.x_timer = ((self.api.width * 16) - x_offset) - self.scores.width * 3
        self.y_timer = y_offset
        self.marked = marked

    def game_reset(self):
        self.__init__(self.api, self.marked, self.x_score, self.y_score)
    
    def draw(self, screen):
        self.draw_score(screen)
        self.draw_timer(screen)

    def draw_timer(self, screen):
        tiles = str(int(self.api.timer)).zfill(3)
        if len(tiles) > 3:
            tiles = "999"
        for x in range(0, 3):
            screen.blit(self.scores[tiles[x]], (self.x_timer + x * self.scores.width, self.y_timer))

    def draw_score(self, screen):
        marked = self.marked()
        if marked < 0:
            marked = 0
        tiles = str(int(marked)).zfill(3)
        if len(tiles) > 3:
            tiles = "999"
        for x in range(0, 3):
            screen.blit(self.scores[tiles[x]], (self.x_score + x * self.scores.width, self.y_score))

    def mouse_up(self, event):
        pass

    def mouse_down(self, event):
        pass

class BoardFace:
    def __init__(self, api, game_reset, x_offset=0, y_offset=8):
        self.api = api
        self.faces = sprites.Face()
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.x = ((api.width * 16) / 2) - self.faces.width / 2 + x_offset
        self.y = y_offset
        self.img = self.faces[':)']
        self.click = None
        self._game_reset = game_reset

    def game_reset(self):
        self.__init__(self.api, self._game_reset, self.x_offset, self.y_offset)

    def draw(self, screen):
        screen.blit(self.img, (self.x, self.y))

    def mouse_up(self, event):
        x, y = event.pos
        if self.mouse_on_face(x, y) and self.mouse_on_face(*self.click):
            self._game_reset()
        elif self.api.is_game_over:
            self.img = self.faces['x(']
        elif self.api.is_game_done:
            self.img = self.faces['B)']
        else:
            self.img = self.faces[':)']
        self.click = None

    def mouse_down_left(self, event):
        x, y = event.pos
        self.click = event.pos
        if self.mouse_on_face(x, y):
            self.click = event.pos
            self.img = self.faces['*:)']
        elif self.api.is_game_over:
            self.img = self.faces['x(']
        elif self.api.is_game_done:
            self.img = self.faces['B)']
        elif self.img == self.faces[':)']:
            self.img = self.faces[':o']
    
    def mouse_on_face(self, x, y):
        return True if (y >= self.y and y <= self.y + self.faces.height) and (x >= self.x and x <= self.x + self.faces.width) else False

    def mouse_down(self, event):
        if event.button == 1:
            self.mouse_down_left(event)

class BoardGame:
    def __init__(self, api, x_offset=0, y_offset=0):
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.api = api
        self.tiles = sprites.Tile(api)
        self.click = None
        self.board = [[self.tiles[self.api.encoder[x]] for x in y] for y in self.api.board]

    def game_reset(self):
        self.api.game_reset()
        self.__init__(self.api, self.x_offset, self.y_offset)
    
    def game_new(self, width, height, mines):
        self.api.game_new(width, height, mines)
        self.__init__(self.api)

    def game_over(self):
        for y, row in enumerate(self.board):
            for x, tile in enumerate(row):
                if self.api.encoder.is_mine(self.api.sheet[y][x]):
                    if tile == self.tiles[self.api.encoder.ascii_tile]:
                        self.board[y][x] = self.tiles[self.api.encoder.ascii_mine]
                else:
                    if tile == self.tiles['f']:
                        self.board[y][x] = self.tiles['fx']
    def game_done(self):
        for y in range(self.api.height):
            for x in range(self.api.width):
                if self.api.encoder.is_tile(self.api.sheet[y][x]):
                    self.board[y][x] = self.tiles['*?']
                else:
                    self.board[y][x] = self.tiles[self.api.encoder.num2ascii(self.api.sheet[y][x])]

    def draw(self, screen):
        for y, row in enumerate(self.board):
            for x, tile in enumerate(row):
                screen.blit(tile, (tile.get_width() * x + self.x_offset, tile.get_height() * y + self.y_offset))

    def mouse_down_left(self, y, x):
        if self.api.is_game_over or not self.api.tile_valid(y, x):
            return
        
        self.click = (y, x)
        if self.board[y][x] == self.tiles[self.api.encoder.ascii_tile]:
            self.board[y][x] = self.tiles['0']
        if self.board[y][x] == self.tiles['?']:
            self.board[y][x] = self.tiles['*?']

    def mouse_down_right(self, y, x):
        if self.api.is_game_over or not self.api.tile_valid(y, x):
            return
        
        if self.board[y][x] == self.tiles[self.api.encoder.ascii_tile]:
            self.board[y][x] = self.tiles['f']
        elif self.board[y][x] == self.tiles['f']:
            self.board[y][x] = self.tiles['?']
        elif self.board[y][x] == self.tiles['?']:
            self.board[y][x] = self.tiles[self.api.encoder.ascii_tile]
    
    def mouse_up_left(self, y, x):
        if self.api.is_game_over or self.click is None:
            pass

        elif self.click[0] != y or self.click[1] != x:
            if self.api.encoder.is_tile(self.api.board[self.click[0]][self.click[1]]):
                if self.board[self.click[0]][self.click[1]] == self.tiles['0']:
                    self.board[self.click[0]][self.click[1]] = self.tiles[self.api.encoder.ascii_tile]
                elif self.board[self.click[0]][self.click[1]] == self.tiles['*?']:
                    self.board[self.click[0]][self.click[1]] = self.tiles['?']
        else:
            if self.api.encoder.is_tile(self.api.board[y][x]) and self.board[y][x] == self.tiles['0']:
                for tile in self.api.tile_open(y, x):
                    v = self.api.encoder.num2ascii(tile['value'])
                    self.board[tile['y']][tile['x']] = self.tiles[v]
                    if self.api.encoder.is_mine(v) and self.api.is_game_over:
                        self.board[y][x] = self.tiles['*x']
                        self.game_over()
            elif self.board[y][x] == self.tiles['*?']:
                self.board[y][x] = self.tiles[self.api.encoder.ascii_tile]
        if self.api.is_game_done:
            self.game_done()
        
        self.click = None
    
    def marked(self):
        c = 0
        for row in self.board:
            c += row.count(self.tiles['f'])
        return self.api.mines - c
    
    def mouse_up_right(self, y, x):
        if self.api.is_game_done:
            self.game_done()

    def mouse_down(self, event):
        y, x = self.mouse_tile(event.pos)
        if event.button == 1:
            self.mouse_down_left(y, x)
        elif event.button == 3:
            self.mouse_down_right(y, x)
    
    def mouse_up(self, event):
        y, x = self.mouse_tile(event.pos)
        if event.button == 1:
            self.mouse_up_left(y, x)
        elif event.button == 3:
            self.mouse_up_right(y, x)
    
    def mouse_tile(self, pos):
        return ((pos[1] - self.y_offset) // 16, (pos[0] - self.x_offset)// 16)